import threading, platform
import whatap.agent.secure.security_master as secure
import whatap.agent.data.data_pack as data
from whatap.agent.counter.task_extrainfo import TaskExtraINFO
from whatap.agent.counter.task_base import TaskBase
from whatap.agent.counter.task_disk import TaskDisk
from whatap.agent.counter.task_network import TaskNet
from whatap.agent.counter.task_proc import TaskProc
from whatap.agent.counter.task_log import TaskLogEvent
from whatap.agent.counter.task_config import TaskConfig
from whatap.agent.counter.task_filesystem import TaskFileSystem
from whatap.agent.counter.task_tcp import TaskTCP
from whatap.agent.counter.task_hostinfo import TaskHostInfo
from whatap.util.date_util import DateUtil as dateutil
import whatap.util.logging_util as logging_util
import whatap.util.panic_util as panic_util
import whatap.osinfo as osinfo
import whatap.agent.conf.configure as config
import whatap.pack.smbase_pack as smbase_pack
import time
from datetime import datetime, timedelta
import multiprocessing

module_loaded = datetime.now()
last_task_timestamp = 0
last_task_name = ""
hangTask = None

class TimeoutException(Exception):
    pass

class TaskAction(threading.Thread):
    def __init__(self, task, name, lastActTime, timeout, timeoutException):
        threading.Thread.__init__(self)
        self.p = None
        self.os = None
        self.pcode = None
        self.oid = None
        self.now = None
        self.timeout = float(timeout)
        self.done_event = threading.Event()
        self.done_condition = threading.Condition()
        self.hangCheck = False
        self.hangCount = 0

        self.task  = task
        self.name = name
        self.lastActTime = lastActTime

        self.timeoutException = timeoutException
        if timeoutException:
            self.setDaemon(True)

    def set(self, os, pcode, oid, now):
        self.os = os
        self.pcode = pcode
        self.oid = oid
        self.now = now

    def run(self):
        while True:
            with self.done_condition:
                if self.done_event.is_set():
                    self.done_condition.wait()

            if not self.done_event.is_set():
                self.p = self.task.process(self.os, self.pcode, self.oid, self.now)
                self.done_event.set()

    def get(self):
        if not self.timeoutException:
            return self.task.process(self.os, self.pcode, self.oid, self.now)

        if self.hangCheck:
            if not self.done_event.is_set():
                self.hangCount += 1
                raise TimeoutException("Hang Check Exception")
            
        with self.done_condition:
            self.done_event.clear()
            self.done_condition.notify()

        self.done_event.wait(timeout=self.timeout)

        if not self.done_event.is_set():
            self.hangCheck = True
            self.hangCount += 1
            raise TimeoutException("Timeout Exception")
        else:
            self.hangCheck = False
            p = self.p
            self.p = None
            return p
        
        return None



class __CounterManager(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self, name='__CounterManager')
        self.setDaemon(True)

    def run(self):
        global last_task_timestamp
        global hangTask

        tasks = []
        securetasks = []

        secu = secure.GetSecurityMaster()
        os = 0
        system = platform.system().lower()
        if system == "windows":
            os = smbase_pack.OS_WINDOW
        elif system == "linux":
            os = smbase_pack.OS_LINUX
        elif system == "darwin":
            os = smbase_pack.OS_OSX
        elif system == "sunos":
            os = smbase_pack.OS_SUNOS
        elif system == "aix":
            os = smbase_pack.OS_AIX
        elif system == "hp-ux":
            os = smbase_pack.OS_HPUX
        else:
            return

        taskTimeoutExcept = config.GetConfig().task_timeout_except
        taskTimeout = config.GetConfig().task_timeout
        taskHealthCheckCount = config.GetConfig().task_health_check_count

        if config.GetConfig().task_base:
            base = TaskAction(TaskBase(), "Base", 0, taskTimeout, taskTimeoutExcept)
            tasks.append(base)
            if taskTimeoutExcept and taskTimeout != 0:
                base.start()

        if config.GetConfig().task_disk:
            disk = TaskAction(TaskDisk(), "Disk", 0, taskTimeout, taskTimeoutExcept)
            tasks.append(disk)
            if taskTimeoutExcept and taskTimeout != 0:
                disk.start()

        if config.GetConfig().task_net:
            net = TaskAction(TaskNet(), "Net", 0, taskTimeout, taskTimeoutExcept)
            tasks.append(net)
            if taskTimeoutExcept and taskTimeout != 0:
                net.start()

        if config.GetConfig().task_proc:
            proc = TaskAction(TaskProc(), "Proc", 0, taskTimeout, taskTimeoutExcept)
            tasks.append(proc)
            if taskTimeoutExcept and taskTimeout != 0:
                proc.start()
            
        if config.GetConfig().task_config:
            con = TaskAction(TaskConfig(), "Config", 0, taskTimeout, taskTimeoutExcept)
            tasks.append(con)
            if taskTimeoutExcept and taskTimeout != 0:
                con.start()

        if config.GetConfig().task_filesystem:
            filesystem = TaskAction(TaskFileSystem(), "Filesystem", 0, taskTimeout, taskTimeoutExcept)
            tasks.append(filesystem)
            if taskTimeoutExcept and taskTimeout != 0:
                filesystem.start()

        if config.GetConfig().task_tcpcheck:
            tcpcheck = TaskAction(TaskTCP(), "TCPCheck", 0, taskTimeout, taskTimeoutExcept)
            tasks.append(tcpcheck)
            if taskTimeoutExcept and taskTimeout != 0:
                tcpcheck.start()

        if config.GetConfig().task_logevent:
            logevent = TaskAction(TaskLogEvent(), "LogEvent", 0, taskTimeout, taskTimeoutExcept)
            securetasks.append(logevent)
            if taskTimeoutExcept and taskTimeout != 0:
                logevent.start()

        if config.GetConfig().task_osinfo:
            info = TaskAction(TaskExtraINFO(), "OSInfo", 0, taskTimeout, taskTimeoutExcept)
            securetasks.append(info)
            if taskTimeoutExcept and taskTimeout != 0:
                info.start()

        if config.GetConfig().HostInfoV2Enabled:
            hostinfo = TaskAction(TaskHostInfo(), "HostInfo", 0, taskTimeout, taskTimeoutExcept)
            securetasks.append(hostinfo)
            if taskTimeoutExcept and taskTimeout != 0:
                hostinfo.start()

        while True:
            try:
                time.sleep(0.5)
                now = osinfo.GetUptime()
                now_timestamp = dateutil.now()
                for each_task in tasks:
                    if each_task.task.interval() >= 5 and now >= each_task.lastActTime+each_task.task.interval() :
                        each_task.lastActTime = now
                        try:
                            logging_util.debug("[CounterManager] TaskStart : ", each_task.name)
                            last_task_name = each_task.name
                            
                            each_task.set(os, secu.pcode, secu.oid, now_timestamp)
                            p = each_task.get()

                            if p :
                                logging_util.debug("[CounterManager] TaskSuccess : ", each_task.name)
                                data.SendHide(p)
                                last_task_timestamp = datetime.now()
                        except TimeoutException, e:
                            if each_task.hangCount >= taskHealthCheckCount:
                                hangTask = each_task.name
                            message = "[CounterManager] Task : {}, Error : {}".format(each_task.name, e)
                            logging_util.error(message)
                            panic_util.AgentErrorLogEventAsync(message)
                        except Exception, e:
                            logging_util.logStack(e)

                for each_task in securetasks:
                    if each_task.task.interval() >= 5 and now >= each_task.lastActTime + each_task.task.interval():
                        each_task.lastActTime = now
                        try:
                            logging_util.debug("[CounterManager] TaskStart : ", each_task.name)
                            last_task_name = each_task.name

                            each_task.set(os, secu.pcode, secu.oid, now_timestamp)
                            p = each_task.get()

                            if p :
                                logging_util.debug("[CounterManager] TaskSuccess : ", each_task.name)
                                data.SendSecure(p)
                                last_task_timestamp = datetime.now()
                        except TimeoutException, e:
                            if each_task.hangCount >= taskHealthCheckCount:
                                hangTask = each_task.name
                            message = "[CounterManager] Task : {}, Error : {}".format(each_task.name, e)
                            logging_util.error(message)
                            panic_util.AgentErrorLogEventAsync(message)
                        except Exception, e:
                            logging_util.logStack(e)
            except Exception, e:
                logging_util.logStack(e)

            
counterManager =__CounterManager()

def startCounterManager():
    counterManager.start()

def isOK():
    global last_task_timestamp
    global hangTask
    ret = True
    now = datetime.now()
    if now-module_loaded > timedelta(minutes = config.GetConfig().HealthCheckCoolTime):
        if last_task_timestamp and now-last_task_timestamp > timedelta(minutes = 1):
            ret = ret and False
            message = "[CounterManager] Health Check Fail Timeout - Last Task : {}".format(last_task_name)
            logging_util.error(message)
            panic_util.AgentErrorLogEventAsync(message)
        if hangTask:
            ret = ret and False
            message = "[CounterManager] Health Check Fail Task Hang : {}".format(hangTask)
            logging_util.error(message)
            panic_util.AgentErrorLogEventAsync(message)

    return ret
